import asyncio
import time
from collections.abc import MutableMapping
from signal import Signals
from typing import Any, Dict, Iterable, Optional, Callable

from amqpworker.conf import logger
from amqpworker.connections import Connection, ConnectionsMapping
from amqpworker.exceptions import InvalidConnection, InvalidRoute
from amqpworker.options import Options, DefaultValues, RouteTypes
from amqpworker.rabbitmq.entrypoints import AMQPRouteEntryPointImpl
from amqpworker.routes import RoutesRegistry, Route
from amqpworker.scheduled import ScheduledThreadPoolExecutor

from amqpworker.signals.base import Freezable, Signal
from concurrent.futures import ThreadPoolExecutor, Executor

from amqpworker.signals.handlers.rabbitmq import RabbitMQ
from amqpworker.utils import entrypoint


class App(MutableMapping, Freezable):
    shutdown_os_signals = (Signals.SIGINT, Signals.SIGTERM)
    handlers = (RabbitMQ(),)

    def __init__(self, connections: Optional[Iterable[Connection]] = None, loop: Optional[Executor] = None):
        Freezable.__init__(self)
        self.routes_registry = RoutesRegistry()
        self.default_route_options: dict = {}
        self.loop: Optional[Executor] = loop
        self.scheduler = None
        self._state: Dict[Any, Any] = self._get_initial_state()
        self.connections = ConnectionsMapping()
        if connections:
            self.connections.add(connections)

        self._on_startup: Signal = Signal(self)
        self._on_shutdown: Signal = Signal(self)

        for handler in self.handlers:
            self._on_startup.append(handler.startup)
            self._on_shutdown.append(handler.shutdown)

        # for signal_ in self.shutdown_os_signals:
        #     signal.signal(signal_, self.shutdown)
        # self.loop.add_signal_handler(signal, self.shutdown)

    def _check_frozen(self):
        if self.frozen:
            raise RuntimeError(
                "You shouldn't change the state of a started application"
            )

    def freeze(self):
        self.connections.freeze()
        super(App, self).freeze()

    def get_connection(self, name: str) -> Connection:
        try:
            return self.connections[name]
        except KeyError as e:
            raise InvalidConnection(
                f"There's no Connection with name `{name}` registered "
                f"in `App.connections`"
            ) from e

    def _get_initial_state(self) -> Dict[str, Dict]:
        # fixme: typeignore reason - https://github.com/python/mypy/issues/4537
        return {route_type: {} for route_type in RouteTypes}  # type: ignore

    def __getitem__(self, key):
        return self._state[key]

    def __setitem__(self, key, value):
        self._check_frozen()
        self._state[key] = value

    def __delitem__(self, key):
        self._check_frozen()
        del self._state[key]

    def __len__(self):
        return len(self._state)

    def __iter__(self):
        return iter(self._state)

    @entrypoint
    def run(self):
        logger.debug({"event": "Booting App..."})
        self.startup()
        try:
            while True:
                time.sleep(10)
        except KeyboardInterrupt:
            self.shutdown()

    def startup(self):
        """
        Causes on_startup signal

        Should be called in the event loop along with the request handler.
        """
        if self.loop is None:
            self.loop = ThreadPoolExecutor(thread_name_prefix='AMQPWORKER-',max_workers=min(16, sum(1 for _ in self.routes_registry.amqp_routes)))
        self._on_startup.send(self)

    def shutdown(self):
        """
        Schedules an on_startup signal

        Is called automatically when the application receives
        a SIGINT or SIGTERM
        """

        return self._on_shutdown.send(self)

    @property
    def amqp(self) -> AMQPRouteEntryPointImpl:
        return AMQPRouteEntryPointImpl(self)

    def route(
            self,
            routes: Iterable[str],
            type: RouteTypes,
            options: Optional[dict] = None,
            **kwargs,
    ):
        if options is None:
            options = {}
        if not isinstance(type, RouteTypes):
            raise TypeError(
                f"type parameter is not a valid RouteTypes." f" Found: '{type}'"
            )

        def wrapper(f):
            handler = f
            if not callable(handler):
                raise TypeError(
                    f"Object passed as handler is not callable: {f}"
                )
            if asyncio.iscoroutinefunction(handler) or (
                    hasattr(handler, '__call__') and asyncio.iscoroutinefunction(handler.__call__)):
                raise TypeError(
                    f"handler must be not a coroutine: {handler}"
                )
            self.routes_registry[handler] = {
                "type": type,
                "routes": routes,
                "handler": handler,
                "options": options,
                "default_options": self.default_route_options,
                **kwargs,
            }
            return f

        return wrapper

    def run_on_startup(self, func: Callable[["App"], None]) -> None:
        """
        Registers a coroutine to be awaited for during app startup
        """
        self._on_startup.append(func)

    def run_on_shutdown(self, func: Callable[["App"], None]) -> None:
        """
        Registers a coroutine to be awaited for during app shutdown
        """
        self._on_shutdown.append(func)

    def run_every(self, seconds: int, options: Optional[Dict] = None):
        """
        Registers a coroutine to be called with a given interval
        """
        if options is None:
            options = {}

        max_concurrency = options.get(
            Options.MAX_CONCURRENCY, DefaultValues.RUN_EVERY_MAX_CONCURRENCY
        )

        def wrapper(task: Callable[..., None]):
            runner = ScheduledThreadPoolExecutor(max_workers=max_concurrency)
            runner.schedule_at_fixed_delay(task, initial_delay=seconds, period=seconds)
            # runner = ScheduledTaskRunner(
            #     seconds=seconds,
            #     task=task,
            #     app=self,
            #     max_concurrency=max_concurrency,
            # )
            self._on_startup.append(runner.start)
            self._on_shutdown.append(runner.stop)
            if "task_runners" not in self:
                self["task_runners"] = []
            self["task_runners"].append(runner)

            return task

        return wrapper

    def get_connection_for_route(self, route_info: Route):
        route_connection = route_info.connection
        if route_connection is None:
            route_connection = route_info.options.get("connection")

        connections = self.connections.with_type(route_info.type)
        if route_connection is not None:
            if isinstance(route_connection, Connection):
                return route_connection
            elif isinstance(route_connection, str):
                return self.get_connection(name=route_connection)
            else:
                # pragma: nocover
                raise InvalidRoute(
                    f"Expected `Route.connection` to be either `str` or "
                    f"`Connection`. Got `{type(route_connection)}` instead."
                )
        elif len(connections) > 1:
            raise InvalidRoute(
                f"Invalid route definition for App. You are trying to "
                f"define a {route_info.type} into an amqpworker.App "
                f"with multiple connections without specifying which "
                f"one to use."
            )
        else:
            try:
                return connections[0]
            except IndexError as e:
                raise InvalidRoute(
                    f"Invalid route definition for App. You are trying to "
                    f"define a {route_info.type} without an "
                    f"Connection registered on App"
                ) from e



class EmptyQueueException(Exception):
    """No message to get"""


class MessageError(ValueError):
    """Base for all message exceptions"""


class UndecodableMessageException(MessageError):
    """Can't decode as JSON"""


class InvalidMessageSizeException(MessageError):
    def __init__(self, message=None):
        """Message size if bigger than it should be"""
        self.message = message



from typing import Callable, Generic, Optional, TypeVar

from amqpstorm import Channel
from amqpworker.easyqueue.connection import AMQPConnection
from amqpworker.easyqueue.exceptions import UndecodableMessageException

T = TypeVar("T")


class AMQPMessage(Generic[T]):
    __slots__ = (
        "connection",
        "channel",
        "queue_name",
        "serialized_data",
        "delivery_tag",
        "_properties",
        "_deserialization_method",
        "_deserialized_data",
        "_queue",
    )

    def __init__(
            self,
            connection: AMQPConnection,
            channel: Channel,
            queue_name: str,
            serialized_data: bytes,
            delivery_tag: int,
            properties: dict,
            deserialization_method: Callable[[bytes], T],
            queue,
    ) -> None:
        self.queue_name = queue_name
        self.serialized_data = serialized_data
        self.delivery_tag = delivery_tag
        self.connection = connection
        self.channel = channel
        self._properties = properties
        self._deserialization_method = deserialization_method

        self._deserialized_data: Optional[T] = None
        self._queue = queue

    @property
    def deserialized_data(self) -> T:
        if self._deserialized_data:
            return self._deserialized_data
        try:
            self._deserialized_data = self._deserialization_method(
                self.serialized_data
            )
        except ValueError as e:
            raise UndecodableMessageException(
                "msg couldn't be decoded as JSON"
            ) from e
        return self._deserialized_data

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        for attr in self.__slots__:
            if attr.startswith("__"):
                continue
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def ack(self):
        return self.channel.basic.ack(self.delivery_tag)

    def reject(self, requeue=False):
        return self.channel.basic.reject(
            delivery_tag=self.delivery_tag, requeue=requeue
        )

    def requeue(self):
        self.channel.basic.ack(self.delivery_tag)

        self.channel.basic.publish(body=self.serialized_data,
                                   routing_key=self.queue_name,
                                   properties=self._properties,
                                   mandatory=False,
                                   immediate=False)



from typing import Any, Dict, Optional, Tuple, Type

from amqpworker.typing import get_args, get_origin


class RegistryItem:
    def __init__(
        self, type: Type, value: Any, type_args: Optional[Tuple] = None
    ) -> None:
        self.type = type
        self.value = value
        self.type_args = type_args


class TypesRegistry:
    def __init__(self):
        self._data: Dict[Tuple, RegistryItem] = {}
        self._by_name: Dict[str, RegistryItem] = {}

    def set(
        self,
        obj: Any,
        type_definition: Optional[Type] = None,
        param_name: Optional[str] = None,
    ) -> None:
        has_type_args = get_args(type_definition)
        origin = get_origin(obj) or obj.__class__

        registry_item = RegistryItem(
            type=origin, value=obj, type_args=has_type_args
        )

        self._data[(origin, has_type_args)] = registry_item

        if param_name:
            self._by_name[param_name] = registry_item

    def get(
        self, _type: Type, param_name: Optional[str] = None, type_args=None
    ) -> Optional[Any]:
        origin = get_origin(_type) or _type
        _type_args = type_args or get_args(_type)
        if param_name:
            try:
                item = self._by_name[param_name]
                if (item.type, item.type_args) == (origin, _type_args):
                    return item.value
            except KeyError:
                return None

        try:
            return self._data[(origin, get_args(_type))].value
        except KeyError:
            return None



from typing import Optional, Tuple, Type, get_type_hints


def get_args(_type: Optional[Type]) -> Optional[Tuple]:
    if _type and hasattr(_type, "__args__"):
        return _type.__args__

    return None


def get_origin(_type: Optional[Type]) -> Optional[Type]:
    if _type and hasattr(_type, "__origin__"):
        return _type.__origin__

    return None


def get_handler_original_typehints(handler):
    """
    Retorna a assinatura do handler amqpworker que está sendo decorado.
    O retorno dessa chamada é equivalente a:
    typing.get_type_hints(original_handler)
    Onde `original_handler` é o handler amqpworker original.

    Ideal para ser usado na pilha de decorators de um handler, ex:

    .. code-block:: python

        @deco1
        @deco2
        @deco3
        async def handler(...):
            pass

    Nesse caso, se qualquer um dos 3 decorators precisar saber a assinatura
    original, deve usar essa função passando a função recebida do decorator anterior.

    """

    def _dummy():
        pass

    _dummy.__annotations__ = getattr(
        handler, "amqpworker_original_annotations", handler.__annotations__
    )

    return get_type_hints(_dummy)


def is_base_type(_type, base_type):
    """
    Retorna True para argumentos de um tipo base `base_type`.
    Ex:
    (a: MyGeneric[int]) -> True
    (b: MyGeneric) -> True
    """
    if get_origin(_type) is base_type:
        return True

    return issubclass(_type, base_type)


def get_handler_original_qualname(handler):
    return getattr(
        handler, "amqpworker_original_qualname", handler.__qualname__
    )



from abc import ABC
from asyncio import iscoroutinefunction
from typing import Generic, TypeVar

import amqpworker
from amqpworker.routes import RouteHandler

T = TypeVar("T")


def _extract_sync_callable(handler) -> RouteHandler:
    cb = handler
    if not callable(cb):
        raise TypeError(f"Object passed as handler is not callable: {cb}")
    if iscoroutinefunction(cb) or (hasattr(cb, '__call__') and iscoroutinefunction(cb)):
        raise TypeError(f"handler must be sync: {cb}")
    return cb


class EntrypointInterface(ABC, Generic[T]):
    def __init__(self, app: "amqpworker.App") -> None:
        self.app = app



import abc
import json
import logging
import time
import traceback
from concurrent.futures import ThreadPoolExecutor
from enum import Enum, auto
from functools import wraps
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Optional,
    Type,
    TypeVar,
    Union,
)

from amqpstorm import Message
from amqpworker.easyqueue.connection import AMQPConnection
from amqpworker.easyqueue.exceptions import UndecodableMessageException
from amqpworker.easyqueue.message import AMQPMessage


class DeliveryModes:
    NON_PERSISTENT = 1
    PERSISTENT = 2


class ConnType(Enum):
    CONSUME = auto()
    WRITE = auto()


class BaseQueue(metaclass=abc.ABCMeta):
    def __init__(
            self,
            host: str,
            username: str,
            password: str,
            port: int = 5672,
            virtual_host: str = "/",
            heartbeat: int = 60,
    ) -> None:
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.virtual_host = virtual_host
        self.heartbeat = heartbeat

    @abc.abstractmethod
    def serialize(self, body: Any, **kwargs) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def deserialize(self, body: bytes) -> Any:
        raise NotImplementedError

    def _parse_message(self, content) -> Dict[str, Any]:
        """
        Gets the raw message body as an input, handles deserialization and
        outputs
        :param content: The raw message body
        """
        try:
            return self.deserialize(content)
        except TypeError:
            return self.deserialize(content.decode())
        except json.decoder.JSONDecodeError as e:
            raise UndecodableMessageException(
                '"{content}" can\'t be decoded as JSON'.format(content=content)
            )


class BaseJsonQueue(BaseQueue):
    content_type = "application/json"

    def serialize(self, body: Any, **kwargs) -> str:
        return json.dumps(body, **kwargs)

    def deserialize(self, body: bytes) -> Any:
        return json.loads(body.decode())


def _ensure_conn_is_ready(conn_type: ConnType):
    def _ensure_connected(coro: Callable[..., Any]):
        @wraps(coro)
        def wrapper(self: "JsonQueue", *args, **kwargs):
            conn = self.conn_types[conn_type]
            retries = 0
            while self.is_running and not conn.has_channel_ready():
                try:
                    conn._connect()
                    break
                except Exception as e:
                    time.sleep(self.seconds_between_conn_retry)
                    retries += 1
                    if self.connection_fail_callback:
                        self.connection_fail_callback(e, retries)
                    if self.logger:
                        self.logger.error(
                            {
                                "event": "reconnect-failure",
                                "retry_count": retries,
                                "exc_traceback": traceback.format_tb(
                                    e.__traceback__
                                ),
                            }
                        )
            return coro(self, *args, **kwargs)

        return wrapper

    return _ensure_connected


T = TypeVar("T")


class _ConsumptionHandler:
    def __init__(
            self,
            delegate: "QueueConsumerDelegate",
            queue: "JsonQueue",
            queue_name: str,
    ) -> None:
        self.delegate = delegate
        self.queue = queue
        self.queue_name = queue_name
        self.consumer_tag: Optional[str] = None

    def _handle_callback(self, callback, **kwargs):
        """
        Chains the callback coroutine into a try/except and calls
        `on_message_handle_error` in case of failure, avoiding unhandled
        exceptions.

        :param callback:
        :param kwargs:
        :return:
        """
        try:
            return callback(**kwargs)
        except Exception as e:
            return self.delegate.on_message_handle_error(
                handler_error=e, **kwargs
            )

    def handle_message(
            self,
            message: Message
    ):

        msg = AMQPMessage(
            connection=self.queue.connection,
            channel=message.channel,
            queue=self.queue,
            properties=message.properties,
            delivery_tag=message.delivery_tag,
            deserialization_method=self.queue.deserialize,
            queue_name=self.queue_name,
            serialized_data=message.body,
        )

        callback = self._handle_callback(
            self.delegate.on_queue_message, msg=msg  # type: ignore
        )
        return callback


class JsonQueue(BaseQueue, Generic[T]):

    def __init__(
            self,
            host: str,
            username: str,
            password: str,
            port: int = 5672,
            delegate_class: Optional[Type["QueueConsumerDelegate"]] = None,
            delegate: Optional["QueueConsumerDelegate"] = None,
            virtual_host: str = "/",
            heartbeat: int = 60,
            prefetch_count: int = 100,
            seconds_between_conn_retry: int = 1,
            logger: Optional[logging.Logger] = None,
            connection_fail_callback: Optional[
                Callable[[Exception, int], None]
            ] = None,
    ) -> None:
        super().__init__(host, username, password, port, virtual_host, heartbeat)

        if delegate is not None and delegate_class is not None:
            raise ValueError("Cant provide both delegate and delegate_class")

        if delegate_class is not None:
            self.delegate = delegate_class()
        else:
            self.delegate = delegate  # type: ignore

        self.prefetch_count = prefetch_count

        on_error = self.delegate.on_connection_error if self.delegate else None

        self.connection = AMQPConnection(
            host=host,
            port=port,
            username=username,
            password=password,
            virtual_host=virtual_host,
            heartbeat=heartbeat,
            on_error=on_error,
        )

        self._write_connection = AMQPConnection(
            host=host,
            port=port,
            username=username,
            password=password,
            virtual_host=virtual_host,
            heartbeat=heartbeat,
            on_error=on_error,
        )

        self.conn_types = {
            ConnType.CONSUME: self.connection,
            ConnType.WRITE: self._write_connection,
        }

        self.seconds_between_conn_retry = seconds_between_conn_retry
        self.is_running = True
        self.logger = logger
        self.connection_fail_callback = connection_fail_callback

    def serialize(self, body: T, **kwargs) -> str:
        return json.dumps(body, **kwargs)

    def deserialize(self, body: bytes) -> T:
        return json.loads(body.decode())

    def conn_for(self, type: ConnType) -> AMQPConnection:
        return self.conn_types[type]

    @_ensure_conn_is_ready(ConnType.WRITE)
    def put(
            self,
            routing_key: str,
            data: Any = None,
            serialized_data: Union[str, bytes] = "",
            exchange: str = "",
            properties: Optional[dict] = None,
            mandatory: bool = False,
            immediate: bool = False,
    ):
        """
        :param data: A serializable data that should be serialized before
        publishing
        :param serialized_data: A payload to be published as is
        :param exchange: The exchange to publish the message
        :param routing_key: The routing key to publish the message
        """

        if data and serialized_data:
            raise ValueError("Only one of data or json should be specified")

        if data:
            if isinstance(data, (str, bytes)):
                serialized_data = data
            else:
                serialized_data = self.serialize(data, ensure_ascii=False)
                if properties is None:
                    properties = {'Content-Type': 'application/json'}
                elif not properties.get('Content-Type'):
                    properties['Content-Type'] = 'application/json'

        if not isinstance(serialized_data, bytes):
            serialized_data = serialized_data.encode()

        return self._write_connection.channel.basic.publish(
            body=serialized_data,
            exchange=exchange,
            routing_key=routing_key,
            properties=properties,
            mandatory=mandatory,
            immediate=immediate,
        )

    @_ensure_conn_is_ready(ConnType.CONSUME)
    def consume(
            self,
            queue_name: str,
            pool: ThreadPoolExecutor,
            delegate: "QueueConsumerDelegate",
            consumer_name: str = "",

    ) -> str:
        """
        Connects the client if needed and starts queue consumption, sending
        `on_before_start_consumption` and `on_consumption_start` notifications
        to the delegate object

        :param queue_name: queue name to consume from
        :param consumer_name: An optional name to be used as a consumer
        identifier. If one isn't provided, a random one is generated by the
        broker
        :return: The consumer tag. Useful for cancelling/stopping consumption
        """
        # todo: Implement a consumer tag generator
        handler = _ConsumptionHandler(
            delegate=delegate, queue=self, queue_name=queue_name
        )

        delegate.on_before_start_consumption(
            queue_name=queue_name, queue=self
        )
        self.connection.channel.basic.qos(
            prefetch_count=self.prefetch_count,
            prefetch_size=0,
            global_=False,
        )
        consumer_tag = self.connection.channel.basic.consume(
            callback=handler.handle_message,
            consumer_tag=consumer_name,
            queue=queue_name,
        )
        delegate.on_consumption_start(
            consumer_tag=consumer_tag, queue=self
        )

        def start():
            try:
                self.connection.channel.start_consuming()
            except:
                traceback.print_exc()

        pool.submit(start)
        handler.consumer_tag = consumer_tag
        return consumer_tag


class QueueConsumerDelegate(metaclass=abc.ABCMeta):
    def on_before_start_consumption(
            self, queue_name: str, queue: JsonQueue
    ):
        """
        Coroutine called before queue consumption starts. May be overwritten to
        implement further custom initialization.

        :param queue_name: Queue name that will be consumed
        :type queue_name: str
        :param queue: AsynQueue instanced
        :type queue: JsonQueue
        """
        pass

    def on_consumption_start(self, consumer_tag: str, queue: JsonQueue):
        """
        Coroutine called once consumption started.
        """

    @abc.abstractmethod
    def on_queue_message(self, msg: AMQPMessage[Any]):
        """
        Callback called every time that a new, valid and deserialized message
        is ready to be handled.

        :param msg: the consumed message
        """
        raise NotImplementedError

    def on_message_handle_error(self, handler_error: Exception, **kwargs):
        """
        Callback called when an uncaught exception was raised during message
        handling stage.

        :param handler_error: The exception that triggered
        :param kwargs: arguments used to call the coroutine that handled
        the message
        :return:
        """
        pass

    def on_connection_error(self, exception: Exception):
        """
        Called when the connection fails
        """
        pass



def wraps(original_handler):
    """
    Esse decorator faz com que a assinatura da função original
    "suba" até o último decorator, que deverá ser sempre um registrador do
    próprio amqpworker. ex:
    @app.http.get(...)
    @deco1
    @deco2
    async def handler(...)
        pass

    Nesse caso, os decorators `@deco1` e `@deco2` devem, *necessariamente*
    fazer uso desse `@wraps()`
    """

    def _wrap(deco):
        deco.amqpworker_original_annotations = getattr(
            original_handler,
            "amqpworker_original_annotations",
            original_handler.__annotations__,
        )
        deco.amqpworker_original_qualname = getattr(
            original_handler,
            "amqpworker_original_qualname",
            original_handler.__qualname__,
        )
        return deco

    return _wrap



from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from amqpworker import App  # noqa: F401


class SignalHandler:
    def startup(self, app: "App"):
        pass  # pragma: no cover

    def shutdown(self, app: "App"):
        pass  # pragma: no cover


"""The MIT License (MIT)

Copyright (c) 2021 Yogaraj.S

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

import threading
import traceback
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, Callable

from delayedqueue import DelayedQueue
from loguru import logger

"""Wrapper for the task submitted to ScheduledThreadPoolExecutor class"""

import time
from typing import Callable


class ScheduledTask:
    def __init__(self, runnable: Callable, initial_delay: int, period: int, *args, time_func=time.time, **kwargs):
        super().__init__()
        self.runnable = runnable
        self.initial_delay = initial_delay
        self.period = period
        self.__time_func = time_func
        self.args = args
        self.kwargs = kwargs
        self.__is_initial = True
        self.task_time = int(self.__time_func() * 1000) + (initial_delay * 1000)

    @property
    def is_initial_run(self) -> bool:
        return self.__is_initial

    @property
    def at_fixed_delay(self) -> bool:
        return self.kwargs.get('is_fixed_delay', False)

    @property
    def at_fixed_rate(self) -> bool:
        return self.kwargs.get('is_fixed_rate', False)

    @property
    def executor_ctx(self):
        return self.kwargs['executor_ctx']  # pragma: no cover

    @property
    def exception_callback(self):
        return self.kwargs.get('on_exception_callback')

    @property
    def time_func(self):
        return self.__time_func

    def __get_next_run(self) -> int:
        if not (self.at_fixed_rate or self.at_fixed_delay):
            raise TypeError("`get_next_run` invoked in a non-repeatable task")
        return int(self.__time_func() * 1000) + self.period * 1000

    def set_next_run(self, time_taken: float = 0) -> None:
        self.__is_initial = False
        self.task_time = self.__get_next_run() - time_taken

    def __lt__(self, other) -> bool:
        if not isinstance(other, ScheduledTask):
            raise TypeError(f"{other} is not of type ScheduledTask")
        return self.task_time < other.task_time

    def __repr__(self) -> str:
        return f"""(Task: {self.runnable.__name__}, Initial Delay: {self.initial_delay} second(s), Periodic: {self.period} second(s), Next run: {time.ctime(self.task_time / 1000)})"""

    def run(self, *args, **kwargs):
        st_time = time.time_ns()
        try:
            self.runnable(*self.args, **self.kwargs)
        except Exception as e:
            if self.exception_callback:
                self.exception_callback(e, *self.args, **self.kwargs)
            else:
                traceback.print_exc()
        finally:
            end_time = time.time_ns()
            time_taken = (end_time - st_time) / 1000000  # in milliseconds
            if self.at_fixed_rate:
                self.set_next_run(time_taken)
                next_delay = (self.period * 1000 - time_taken) / 1000
                if next_delay < 0 or self.task_time <= (self.__time_func() * 1000):
                    self.executor_ctx._put(self, 0)
                else:
                    self.executor_ctx._put(self, next_delay)
            elif self.at_fixed_delay:
                self.set_next_run()
                self.executor_ctx._put(self, self.period)


class ScheduledThreadPoolExecutor(ThreadPoolExecutor):
    def __init__(self, max_workers=10, name=''):
        super().__init__(max_workers=max_workers, thread_name_prefix=name)
        self._max_workers = max_workers
        self.queue = DelayedQueue()
        self.shutdown = False

    def schedule_at_fixed_rate(self, fn: Callable, initial_delay: int, period: int, *args, **kwargs) -> bool:
        if self.shutdown:
            raise RuntimeError(f"cannot schedule new task after shutdown!")
        task = ScheduledTask(fn, initial_delay, period, *args, is_fixed_rate=True, executor_ctx=self, **kwargs)
        return self._put(task, initial_delay)

    def schedule_at_fixed_delay(self, fn: Callable, initial_delay: int, period: int, *args, **kwargs) -> bool:
        if self.shutdown:
            raise RuntimeError(f"cannot schedule new task after shutdown!")
        task = ScheduledTask(fn, initial_delay, period, *args, is_fixed_delay=True, executor_ctx=self, **kwargs)
        return self._put(task, initial_delay)

    def schedule(self, fn, initial_delay, *args, **kwargs) -> bool:
        task = ScheduledTask(fn, initial_delay, 0, *args, executor_ctx=self, **kwargs)
        return self._put(task, initial_delay)

    def _put(self, task: ScheduledTask, delay: int) -> bool:
        # Don't use this explicitly. Use schedule/schedule_at_fixed_delay/schedule_at_fixed_rate. Additionally, to be
        # called by ScheduledTask only!
        if not isinstance(task, ScheduledTask):
            raise TypeError(f"Task `{task!r}` must be of type ScheduledTask")
        if delay < 0:
            raise ValueError(f"Delay `{delay}` must be a non-negative number")
        # logger.debug(f" enqueuing {task} with delay of {delay}")
        return self.queue.put(task, delay)

    def __run(self):
        while not self.shutdown:
            try:
                task: ScheduledTask = self.queue.get()
                super().submit(task.run, *task.args, **task.kwargs)
            except Exception as e:
                print(e)

    def stop(self, app, wait_for_completion: Optional[bool] = False):
        self.shutdown = True
        super().shutdown(wait_for_completion)

    def start(self, app):
        t = threading.Thread(target=self.__run)
        t.setDaemon(True)
        t.start()



from typing import List, Optional

from amqpworker import conf
from amqpworker.connections import AMQPConnection
from amqpworker.entrypoints import EntrypointInterface, _extract_sync_callable
from amqpworker.routes import AMQPRoute, AMQPRouteOptions, RoutesRegistry


def _register_amqp_handler(
    registry: RoutesRegistry,
    routes: List[str],
    vhost: str,
    connection: Optional[AMQPConnection],
    options: Optional[AMQPRouteOptions],
):
    def _wrap(f):
        cb = _extract_sync_callable(f)
        route = AMQPRoute(
            handler=cb,
            routes=routes,
            vhost=vhost,
            connection=connection,
            options=options,
        )
        registry.add_amqp_route(route)

        return f

    return _wrap


class AMQPRouteEntryPointImpl(EntrypointInterface):
    def consume(
        self,
        routes: List[str],
        vhost: str = conf.settings.AMQP_DEFAULT_VHOST,
        connection: Optional[AMQPConnection] = None,
        options: Optional[AMQPRouteOptions] = AMQPRouteOptions(),
    ):
        return _register_amqp_handler(
            self.app.routes_registry, routes, vhost, connection, options
        )
